"""Integration tests for the StringsCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a StringsCommand instance."""
    yield pebble_shell.commands.StringsCommand(shell=shell)


def test_name(command: pebble_shell.commands.StringsCommand):
    assert command.name == "strings"


def test_category(command: pebble_shell.commands.StringsCommand):
    assert command.category == "File Analysis"


def test_help(command: pebble_shell.commands.StringsCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "strings" in output
    assert "Extract printable strings" in output
    assert "binary" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "strings" in output
    assert "Extract printable strings" in output


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # strings with no args should fail and show usage
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should fail with usage message or read from stdin
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["usage", "Usage", "strings", "file"])
    else:
        # May succeed waiting for stdin
        assert result == 0


def test_execute_binary_file_analysis(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test analyzing binary file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should either succeed extracting strings or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain printable strings from binary
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should contain ASCII strings
            lines = output.strip().split("\n")
            for line in lines:
                # Each line should be a printable string
                assert all(
                    ord(char) >= 32 and ord(char) <= 126
                    for char in line
                    if line.strip()
                )
    else:
        # Should fail if file doesn't exist or access denied
        assert result == 1


def test_execute_text_file_analysis(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test analyzing text file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should either succeed extracting strings or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain text from file
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should contain readable text
            lines = output.strip().split("\n")
            assert len(lines) >= 1
    else:
        # Should fail if file doesn't exist or access denied
        assert result == 1


def test_execute_minimum_length_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test minimum length option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "10", "/bin/ls"])

    # Should either succeed with length filtering or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain strings of minimum length
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            lines = output.strip().split("\n")
            for line in lines:
                if line.strip():
                    # Each string should be at least 10 characters
                    assert len(line.strip()) >= 10
    else:
        assert result == 1


def test_execute_all_characters_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test all characters option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a", "/bin/ls"])

    # Should either succeed scanning all sections or fail gracefully
    if result == 0:
        output = capture.get()
        # Should scan entire file
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_print_filename_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test print filename option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-f", "/bin/ls"])

    # Should either succeed with filename printing or fail gracefully
    if result == 0:
        output = capture.get()
        # Should include filename in output
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should contain filename
            assert "/bin/ls" in output or "ls" in output
    else:
        assert result == 1


def test_execute_print_offset_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test print offset option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-o", "/bin/ls"])

    # Should either succeed with offset printing or fail gracefully
    if result == 0:
        output = capture.get()
        # Should include offsets in output
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should contain numeric offsets
            assert any(char.isdigit() for char in output)
    else:
        assert result == 1


def test_execute_radix_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test radix option for offset display
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "x", "/bin/ls"])

    # Should either succeed with hex offsets or fail gracefully
    if result == 0:
        output = capture.get()
        # Should include hex offsets
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # May contain hexadecimal values
            assert any(char in output for char in "0123456789abcdefABCDEF")
    else:
        assert result == 1


def test_execute_encoding_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test encoding option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-e", "s", "/bin/ls"])

    # Should either succeed with specific encoding or fail gracefully
    if result == 0:
        output = capture.get()
        # Should extract strings with specified encoding
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_multiple_files(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test processing multiple files
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls", "/bin/cat"])

    # Should either succeed processing multiple files or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain strings from both files
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test with nonexistent file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent/file.bin"])

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output for msg in ["No such file", "cannot open", "error", "not found"]
    )


def test_execute_permission_denied_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test with permission denied file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/root/.ssh/id_rsa"])

    # Should either succeed or fail with permission error
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "cannot", "error"])
    else:
        # May succeed if file is readable
        assert result == 0


def test_execute_empty_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test with empty file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/dev/null"])

    # Should succeed with no output
    assert result == 0
    output = capture.get()
    # Should produce no strings for empty file
    assert len(output.strip()) == 0


def test_execute_large_binary_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test with large binary file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/usr/bin/bash"])

    # Should either succeed processing large file or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle large file efficiently
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_executable_file_analysis(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test analyzing executable file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/echo"])

    # Should either succeed extracting strings or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain strings from executable
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should contain printable strings
            lines = output.strip().split("\n")
            assert len(lines) >= 1
    else:
        assert result == 1


def test_execute_library_file_analysis(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test analyzing library file
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/lib/x86_64-linux-gnu/libc.so.6"]
        )

    # Should either succeed or fail gracefully if file doesn't exist
    if result == 0:
        output = capture.get()
        # Should contain strings from library
        assert len(output.strip()) >= 0
    else:
        # Library may not exist on all systems
        assert result == 1


def test_execute_short_strings_filtering(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test filtering short strings
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "4", "/bin/ls"])

    # Should either succeed with minimum length 4 or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain strings of length 4 or more
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            lines = output.strip().split("\n")
            for line in lines:
                if line.strip():
                    assert len(line.strip()) >= 4
    else:
        assert result == 1


def test_execute_ascii_strings_only(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test extracting ASCII strings only
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should extract ASCII printable strings
    if result == 0:
        output = capture.get()
        # Should contain only ASCII printable characters
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            for char in output:
                if char not in "\n\r":
                    # Should be printable ASCII
                    assert 32 <= ord(char) <= 126
    else:
        assert result == 1


def test_execute_unicode_string_detection(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test Unicode string detection
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-e", "l", "/bin/ls"])

    # Should either succeed with Unicode detection or fail gracefully
    if result == 0:
        output = capture.get()
        # Should detect Unicode strings if present
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_wide_character_strings(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test wide character string detection
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-e", "b", "/bin/ls"])

    # Should either succeed with wide character detection or fail gracefully
    if result == 0:
        output = capture.get()
        # Should detect wide character strings
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test invalid option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z", "/bin/ls"])

    # Should fail with invalid option error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "option", "usage"])


def test_execute_invalid_minimum_length(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test invalid minimum length
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "invalid", "/bin/ls"])

    # Should fail with invalid length error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "number", "error"])


def test_execute_zero_minimum_length(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test zero minimum length
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "0", "/bin/ls"])

    # Should either succeed or handle zero length appropriately
    if result == 0:
        output = capture.get()
        # Should extract all strings
        assert len(output.strip()) >= 0
    else:
        # May reject zero length
        assert result == 1


def test_execute_very_large_minimum_length(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test very large minimum length
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "1000", "/bin/ls"])

    # Should succeed but find few or no strings
    assert result == 0
    output = capture.get()
    # Should produce minimal output with very long minimum length
    assert len(output.strip()) >= 0


def test_execute_output_format_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test output format validation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should produce properly formatted output
    if result == 0:
        output = capture.get()
        # Should have valid string output
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should be line-oriented output
            lines = output.strip().split("\n")
            assert len(lines) >= 1
    else:
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 100000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_concurrent_execution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test concurrent execution safety
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/echo"])

    # Should handle concurrent execution safely
    if result == 0:
        output = capture.get()
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "abc", "/bin/ls"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["invalid", "error", "number"])


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test signal handling during processing
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should handle signals appropriately
    if result == 0:
        output = capture.get()
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_locale_awareness(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test locale awareness
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should handle locale-specific character sets
    if result == 0:
        output = capture.get()
        # Should process according to locale
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_file_type_detection(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test file type detection
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should handle different file types appropriately
    if result == 0:
        output = capture.get()
        # Should process text files correctly
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_string_extraction_accuracy(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test string extraction accuracy
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/echo"])

    # Should extract strings accurately
    if result == 0:
        output = capture.get()
        # Should contain meaningful strings
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should contain reasonable strings
            lines = output.strip().split("\n")
            assert any(len(line.strip()) >= 3 for line in lines if line.strip())
    else:
        assert result == 1


def test_execute_cross_platform_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test cross-platform compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should work across different platforms
    if result == 0:
        output = capture.get()
        # Should adapt to platform binary formats
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_binary_format_support(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test binary format support
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should support various binary formats
    if result == 0:
        output = capture.get()
        # Should handle ELF, PE, or other formats
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_performance_optimization(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test performance optimization
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should complete efficiently
    if result == 0:
        output = capture.get()
        # Should process efficiently
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_character_encoding_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test character encoding handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should handle various character encodings
    if result == 0:
        output = capture.get()
        # Should decode characters properly
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_data_consistency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.StringsCommand,
):
    # Test data consistency
    with command.shell.console.capture() as _:
        result1 = command.execute(client=client, args=["/bin/echo"])

    with command.shell.console.capture() as _:
        result2 = command.execute(client=client, args=["/bin/echo"])

    # Should produce consistent results
    if result1 == 0 and result2 == 0:
        # Both executions should succeed consistently
        assert result1 == result2
    else:
        # At least one should succeed or both should fail consistently
        assert result1 == result2



"""Implementation of IdCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag
from ...utils.proc_reader import (
    ProcReadError,
    get_group_name_for_gid,
    get_user_name_for_uid,
    read_proc_status_fields,
)
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class IdCommand(Command):
    """Command for displaying user and group IDs."""

    name = "id"
    help = "Show user and group IDs"
    category = "Built-in Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute the id command to show user and group IDs."""
        if handle_help_flag(self, args):
            return 0

        try:
            status_fields = read_proc_status_fields(client, "self", ["Uid", "Gid"])

            # Extract real UID and GID (first value in each field)
            uid = "0"
            gid = "0"

            if "Uid" in status_fields:
                uid_parts = status_fields["Uid"].split()
                if uid_parts:
                    uid = uid_parts[0]

            if "Gid" in status_fields:
                gid_parts = status_fields["Gid"].split()
                if gid_parts:
                    gid = gid_parts[0]

            uid_name = get_user_name_for_uid(client, uid)
            gid_name = get_group_name_for_gid(client, gid)

            uid_str = f"uid={uid}({uid_name})" if uid_name else f"uid={uid}"
            gid_str = f"gid={gid}({gid_name})" if gid_name else f"gid={gid}"

            self.console.print(f"{uid_str} {gid_str}")
            return 0

        except ProcReadError:
            self.console.print("uid=0 gid=0")
            return 1



"""Implementation of StringsCommand."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Union

import ops

from pebble_shell.utils.command_helpers import safe_read_file

from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


# TODO: Put this in the text category.
class StringsCommand(Command):
    """Print printable strings from files."""

    name = "strings"
    help = "Print printable strings from files"
    category = "File"

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the strings command."""
        return self._execute_strings(client, args)

    def _execute_strings(self, client: ClientType, args: list[str]) -> int:
        if not args:
            self.console.print("Usage: strings [files...]")
            return 1

        for file_path in args:
            try:
                content = safe_read_file(client, file_path, self.shell)
                if content is not None:
                    # Find printable strings (4+ consecutive printable chars)
                    strings = re.findall(r"[!-~]{4,}", content)
                    for string in strings:
                        self.console.print(string)
            except Exception as e:  # noqa: PERF203  # needed for robust file processing
                self.console.print(f"strings: {file_path}: {e}")
                return 1

        return 0



"""Integration tests for the Unix2dosCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a Unix2dosCommand instance."""
    yield pebble_shell.commands.Unix2dosCommand(shell=shell)


def test_name(command: pebble_shell.commands.Unix2dosCommand):
    assert command.name == "unix2dos"


def test_category(command: pebble_shell.commands.Unix2dosCommand):
    assert command.category == "Text Processing"


def test_help(command: pebble_shell.commands.Unix2dosCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "unix2dos" in output
    assert any(
        phrase in output.lower()
        for phrase in ["dos", "unix", "line ending", "convert", "crlf"]
    )


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "unix2dos" in output
    assert any(
        phrase in output.lower() for phrase in ["dos", "unix", "line ending", "convert"]
    )


def test_execute_no_args_stdin(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # unix2dos with no args should read from stdin
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either read from stdin or show usage
    if result == 0:
        output = capture.get()
        # Should wait for stdin input
        assert len(output) >= 0
    else:
        # Should fail with usage message
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["usage", "Usage", "unix2dos", "file"])


def test_execute_convert_unix_to_dos_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test converting Unix line endings to DOS
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should either succeed converting or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain file contents with DOS line endings
        assert len(output) >= 0
        # Should contain DOS line endings (CRLF)
        if len(output) > 0 and "\n" in output:
            # Should have converted some LF to CRLF
            line_endings = output.count("\n")
            if line_endings > 0:
                # Should contain CRLF sequences
                assert "\r\n" in output or line_endings > 0
    else:
        # Should fail if file doesn't exist or access denied
        assert result == 1


def test_execute_in_place_conversion(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test in-place file conversion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-k", "/etc/hosts"])

    # Should either succeed with in-place conversion or fail gracefully
    if result == 0:
        output = capture.get()
        # Should convert file in place (may have no output)
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_keep_date_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test keeping original file date
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-k", "/etc/passwd"])

    # Should either succeed keeping date or fail gracefully
    if result == 0:
        output = capture.get()
        # Should preserve file timestamp
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_quiet_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test quiet mode operation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-q", "/etc/hosts"])

    # Should either succeed quietly or fail gracefully
    if result == 0:
        output = capture.get()
        # Should produce minimal output in quiet mode
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_verbose_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test verbose mode operation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-v", "/etc/passwd"])

    # Should either succeed with verbose output or fail gracefully
    if result == 0:
        output = capture.get()
        # Should provide detailed information in verbose mode
        assert len(output) >= 0
        if len(output) > 0:
            # Should contain status information
            assert any(
                word in output.lower()
                for word in ["convert", "process", "file", "line"]
            )
    else:
        assert result == 1


def test_execute_backup_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test creating backup of original file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-b", "/etc/hosts"])

    # Should either succeed creating backup or fail gracefully
    if result == 0:
        output = capture.get()
        # Should create backup file
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_force_conversion(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test forcing conversion of binary files
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-f", "/etc/passwd"])

    # Should either succeed with force or fail gracefully
    if result == 0:
        output = capture.get()
        # Should force conversion
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_new_file_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test new file mode (write to different file)
    with tempfile.NamedTemporaryFile(suffix="_hosts_dos", delete=False) as output_file:
        output_path = output_file.name
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "/etc/hosts", output_path])

    # Should either succeed with new file or fail gracefully
    if result == 0:
        output = capture.get()
        # Should create new file with DOS line endings
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_preserve_ownership(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test preserving file ownership
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-o", "/etc/passwd"])

    # Should either succeed preserving ownership or fail gracefully
    if result == 0:
        output = capture.get()
        # Should preserve file ownership
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_safe_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test safe mode (skip binary files)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/etc/hosts"])

    # Should either succeed in safe mode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should skip binary files safely
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_ascii_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test ASCII mode conversion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-ascii", "/etc/passwd"])

    # Should either succeed in ASCII mode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should convert in ASCII mode
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_iso_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test ISO mode conversion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-iso", "/etc/hosts"])

    # Should either succeed in ISO mode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should convert in ISO mode
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_utf8_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test UTF-8 mode conversion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-utf8", "/etc/passwd"])

    # Should either succeed in UTF-8 mode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should convert in UTF-8 mode
        assert len(output) >= 0
        if len(output) > 0:
            # Should handle UTF-8 encoding properly
            assert isinstance(output, str)
    else:
        assert result == 1


def test_execute_empty_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test with empty file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/dev/null"])

    # Should succeed with no output for empty file
    if result == 0:
        output = capture.get()
        # Should produce no output for empty file
        assert len(output) == 0
    else:
        # May fail if cannot read/write /dev/null
        assert result == 1


def test_execute_file_with_dos_endings(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test file that already has DOS line endings
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hostname"])

    # Should succeed with no changes needed or double conversion
    if result == 0:
        output = capture.get()
        # Should handle existing DOS format appropriately
        assert len(output) >= 0
        if len(output) > 0 and "\n" in output:
            # Should have DOS line endings
            assert "\r\n" in output or output.count("\n") >= 1
    else:
        assert result == 1


def test_execute_file_with_mixed_endings(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test file with mixed line endings
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should succeed normalizing mixed line endings
    if result == 0:
        output = capture.get()
        # Should normalize all line endings to DOS format
        assert len(output) >= 0
        if len(output) > 0 and "\n" in output:
            # Should contain DOS line endings
            line_count = output.count("\n")
            if line_count > 0:
                # Should have CRLF sequences
                assert "\r\n" in output or line_count >= 1
    else:
        assert result == 1


def test_execute_single_line_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test single line file without newline
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hostname"])

    # Should succeed with single line
    if result == 0:
        output = capture.get()
        # Should handle single line appropriately
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_multiple_files(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test processing multiple files
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts", "/etc/hostname"])

    # Should either succeed with multiple files or fail gracefully
    if result == 0:
        output = capture.get()
        # Should process both files
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test with nonexistent file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent/file.txt"])

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output for msg in ["No such file", "cannot", "error", "not found"]
    )


def test_execute_permission_denied_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test with permission denied file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/root/.ssh/id_rsa"])

    # Should either succeed or fail with permission error
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "cannot", "error"])
    else:
        # May succeed if file is readable
        assert result == 0


def test_execute_directory_as_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test with directory instead of file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc"])

    # Should fail with directory error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["directory", "Is a directory", "error"])


def test_execute_binary_file_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test with binary file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should either handle binary or fail gracefully
    if result == 0:
        output = capture.get()
        # Should process binary file (may skip or convert)
        assert len(output) >= 0
    else:
        # May fail for binary files in safe mode
        assert result == 1


def test_execute_readonly_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test with read-only file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should either succeed or fail appropriately for read-only
    if result == 0:
        output = capture.get()
        # Should handle read-only files
        assert len(output) >= 0
    else:
        # May fail if cannot write to read-only file
        assert result == 1


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test invalid option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z", "/etc/hosts"])

    # Should fail with invalid option error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "option", "usage"])


def test_execute_conflicting_options(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test conflicting options
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-ascii", "-utf8", "/etc/hosts"])

    # Should either handle conflicts or fail appropriately
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["conflicting", "invalid", "error"])
    else:
        # May resolve conflicts automatically
        assert result == 0


def test_execute_unicode_content(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test file with Unicode content
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-utf8", "/etc/hosts"])

    # Should either succeed with Unicode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle Unicode characters properly
        assert len(output) >= 0
        # Should maintain Unicode character integrity
        if len(output) > 0:
            assert isinstance(output, str)
    else:
        assert result == 1


def test_execute_large_file_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test with large file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should handle large files efficiently
    if result == 0:
        output = capture.get()
        # Should process large files
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_line_ending_conversion_accuracy(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test accuracy of line ending conversion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should accurately convert line endings
    if result == 0:
        output = capture.get()
        # Should convert Unix LF to DOS CRLF
        assert len(output) >= 0
        if len(output) > 0 and "\n" in output:
            # Should have proper DOS line endings
            lines = output.split("\n")
            if len(lines) > 1:
                # Should contain CRLF sequences
                assert "\r\n" in output or len(lines) >= 2
    else:
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 100000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_performance_optimization(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test performance optimization
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should complete efficiently
    if result == 0:
        output = capture.get()
        # Should process efficiently
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_data_integrity(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test data integrity during conversion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should maintain data integrity
    if result == 0:
        output = capture.get()
        # Should preserve all content except line endings
        assert len(output) >= 0
        if len(output) > 0:
            # Should not lose any data
            content_without_crlf = output.replace("\r\n", "\n").replace("\n", "")
            assert len(content_without_crlf) > 0
    else:
        assert result == 1


def test_execute_consistent_conversion(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test consistent conversion behavior
    with command.shell.console.capture() as _:
        result1 = command.execute(client=client, args=["/etc/hosts"])

    with command.shell.console.capture() as _:
        result2 = command.execute(client=client, args=["/etc/hosts"])

    # Should produce consistent results
    if result1 == 0 and result2 == 0:
        # Both executions should succeed consistently
        assert result1 == result2
    else:
        # At least one should succeed or both should fail consistently
        assert result1 == result2


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z", "/etc/hosts"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["invalid", "unknown", "error", "usage"])


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test signal handling during processing
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should handle signals appropriately
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_cross_platform_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.Unix2dosCommand,
):
    # Test cross-platform compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should work across different platforms
    if result == 0:
        output = capture.get()
        # Should handle platform-specific line endings
        assert len(output) >= 0
        if len(output) > 0 and "\n" in output:
            # Should produce DOS-compatible line endings
            lines = output.count("\n")
            if lines > 0:
                # Should contain appropriate line endings
                assert "\r\n" in output or lines >= 1
    else:
        assert result == 1



"""Implementation of JqCommand."""

from __future__ import annotations

import json
import os
import sys
from typing import TYPE_CHECKING, Any, Union, cast

import ops
from rich.panel import Panel
from rich.syntax import Syntax

from ...utils import resolve_path
from ...utils.command_helpers import handle_help_flag
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class JqCommand(Command):
    """Command for pretty-printing JSON files with optional keypath filtering."""

    name = "jq"
    help = "Pretty-print JSON files with optional jq-like keypath filtering. Usage: jq <file> [.foo.bar]"
    category = "Built-in Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute the jq command to format and filter JSON files."""
        if handle_help_flag(self, args):
            return 0

        if not args:
            self.console.print(
                Panel(
                    "No file specified. Usage: jq <file> [.foo.bar]",
                    title="[b red]jq Error[/b red]",
                    style="red",
                )
            )
            return 1

        file_path = args[0]
        keypath = args[1] if len(args) > 1 else None
        cwd = self.shell.current_directory

        # Resolve path
        if file_path == "-":
            # Read from stdin
            content = sys.stdin.read()
        else:
            if not os.path.isabs(file_path):
                file_path = os.path.normpath(os.path.join(cwd, file_path))
            try:
                resolved_path = resolve_path(cwd, file_path, self.shell.home_dir)
                with client.pull(resolved_path) as file:
                    content = file.read()
                    if isinstance(content, bytes):
                        content = content.decode("utf-8")
            except Exception as e:
                self.console.print(
                    Panel(
                        f"Error reading file {file_path}: {e}",
                        title="[b red]jq Error[/b red]",
                        style="red",
                    )
                )
                return 1

        # Parse JSON
        try:
            data = json.loads(content)
        except Exception as e:
            self.console.print(
                Panel(
                    f"Error parsing JSON: {e}",
                    title="[b red]jq Error[/b red]",
                    style="red",
                )
            )
            return 1

        # Apply keypath filter if provided
        if keypath:
            if keypath.startswith("."):
                keys = [k for k in keypath.lstrip(".").split(".") if k]
                try:
                    for k in keys:
                        if isinstance(data, dict):
                            data = cast("dict[Any, Any]", data)
                            data = data[k]
                        elif isinstance(data, list) and k.isdigit():
                            data = cast("list[Any]", data)
                            data = data[int(k)]
                        else:
                            raise KeyError(k)
                except Exception as e:
                    self.console.print(
                        Panel(
                            f"Key path error: {e}",
                            title="[b red]jq Error[/b red]",
                            style="red",
                        )
                    )
                    return 1
            else:
                self.console.print(
                    Panel(
                        f"Invalid keypath: {keypath}. Must start with a dot (e.g., .foo.bar)",
                        title="[b red]jq Error[/b red]",
                        style="red",
                    )
                )
                return 1

        # Pretty-print JSON with syntax highlighting
        json_str = json.dumps(data, indent=2, ensure_ascii=False)
        syntax = Syntax(
            json_str,
            "json",
            theme="monokai",
            word_wrap=True,
            line_numbers=True,
            background_color="default",
        )
        self.console.print(
            Panel(
                syntax,
                title="[b]jq Output[/b]",
                style="cyan",
                border_style="bright_blue",
            )
        )
        return 0



"""Implementation of ifconfig command (read-only)."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from ...utils.command_helpers import handle_help_flag, safe_read_file
from ...utils.proc_reader import ProcReadError, read_proc_file
from .._base import Command

if TYPE_CHECKING:
    import ops
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union["ops.pebble.Client", "shimmer.PebbleCliClient"]


class IfconfigCommand(Command):
    """Implementation of ifconfig command (read-only)."""

    name = "ifconfig"
    help = "Display network interface configuration"
    category = "Network"

    def show_help(self):
        """Show command help."""
        help_text = """Display network interface configuration.

Usage: ifconfig [INTERFACE]

Description:
    Display configuration of network interfaces. Read-only version.

Options:
    -h, --help      Show this help message

Examples:
    ifconfig        # Show all interfaces
    ifconfig eth0   # Show specific interface
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the ifconfig command."""
        if handle_help_flag(self, args):
            return 0

        interface = args[0] if args else None

        try:
            # Read network interfaces from /proc/net/dev
            try:
                net_dev = read_proc_file(client, "/proc/net/dev")
            except ProcReadError:
                self.console.print(
                    "[red]ifconfig: cannot read network interface information[/red]"
                )
                return 1

            interfaces = {}
            lines = net_dev.strip().split("\n")

            # Skip header lines
            for line in lines[2:]:
                if ":" in line:
                    iface_name, stats = line.split(":", 1)
                    iface_name = iface_name.strip()

                    if interface and iface_name != interface:
                        continue

                    stats_values = stats.split()
                    if len(stats_values) >= 16:
                        interfaces[iface_name] = {
                            "rx_bytes": int(stats_values[0]),
                            "rx_packets": int(stats_values[1]),
                            "rx_errors": int(stats_values[2]),
                            "tx_bytes": int(stats_values[8]),
                            "tx_packets": int(stats_values[9]),
                            "tx_errors": int(stats_values[10]),
                        }

            if interface and interface not in interfaces:
                self.console.print(
                    f"[red]ifconfig: interface '{interface}' not found[/red]"
                )
                return 1

            # Display interface information
            for iface_name, stats in interfaces.items():
                self._display_interface(client, iface_name, stats)
                if len(interfaces) > 1:
                    self.console.print()

            return 0

        except Exception as e:
            self.console.print(f"[red]ifconfig: {e}[/red]")
            return 1

    def _display_interface(self, client: ClientType, iface_name: str, stats: dict):
        """Display information for a single interface."""
        self.console.print(f"[bold]{iface_name}[/bold]:", end="")

        # Try to get additional interface info
        flags = []
        try:
            # Check if interface is up
            operstate_path = f"/sys/class/net/{iface_name}/operstate"
            try:
                operstate = safe_read_file(client, operstate_path, self.shell)
                if operstate and operstate.strip() == "up":
                    flags.append("UP")
                else:
                    flags.append("DOWN")
            except Exception:
                # Broad exception needed for network interface probing
                flags.append("UNKNOWN")

            # Try to get MTU
            try:
                mtu_path = f"/sys/class/net/{iface_name}/mtu"
                mtu = safe_read_file(client, mtu_path, self.shell)
                mtu_value = mtu.strip() if mtu else "1500"
            except Exception:
                # Broad exception needed for network interface probing
                mtu_value = "1500"

            # Try to get MAC address
            try:
                addr_path = f"/sys/class/net/{iface_name}/address"
                mac_addr = safe_read_file(client, addr_path, self.shell)
                mac_addr = mac_addr.strip() if mac_addr else "00:00:00:00:00:00"
            except Exception:
                # Broad exception needed for network interface probing
                mac_addr = "00:00:00:00:00:00"

            flags.append("BROADCAST")
            flags.append("MULTICAST")

            self.console.print(f" flags=<{','.join(flags)}> mtu {mtu_value}")
            self.console.print(f"        ether {mac_addr}")

            # Display statistics
            self.console.print(
                f"        RX packets {stats['rx_packets']} bytes {stats['rx_bytes']} errors {stats['rx_errors']}"
            )
            self.console.print(
                f"        TX packets {stats['tx_packets']} bytes {stats['tx_bytes']} errors {stats['tx_errors']}"
            )

        except Exception:
            # Broad exception needed for network interface discovery
            self.console.print(" flags=<UNKNOWN>")
            self.console.print(
                f"        RX packets {stats['rx_packets']} bytes {stats['rx_bytes']}"
            )
            self.console.print(
                f"        TX packets {stats['tx_packets']} bytes {stats['tx_bytes']}"
            )



"""Integration tests for the SeqCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a SeqCommand instance."""
    yield pebble_shell.commands.SeqCommand(shell=shell)


def test_name(command: pebble_shell.commands.SeqCommand):
    assert command.name == "seq"


def test_category(command: pebble_shell.commands.SeqCommand):
    assert command.category == "Text Utilities"


def test_help(command: pebble_shell.commands.SeqCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "Generate sequences" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SeqCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "Generate sequences" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SeqCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=[])
    assert result == 1
    # Should show usage information


def test_execute_single_number(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SeqCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["5"])
    assert result == 0
    # Should generate sequence 1 to 5
    output = capture.get()
    assert "1" in output and "5" in output


def test_execute_range(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SeqCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["3", "7"])
    assert result == 0
    # Should generate sequence 3 to 7
    output = capture.get()
    assert "3" in output and "7" in output


def test_execute_invalid_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SeqCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["invalid"])
    assert result == 1



"""Standardised error handling utilities for Pebble operations.

This module provides common error handling patterns and utilities that are
frequently used across command modules when working with Pebble client operations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

import ops
from rich.panel import Panel

if TYPE_CHECKING:
    from collections.abc import Callable

    from rich.console import Console

T = TypeVar("T")


def handle_pebble_path_error(
    console: Console, operation: str, path: str, error: ops.pebble.PathError
) -> None:
    """Handle Pebble PathError with standardized formatting.

    Args:
        console: Rich console for output
        operation: Description of the operation that failed (e.g., "read file")
        path: File path that caused the error
        error: The PathError that occurred
    """
    error_message = f"Failed to {operation}: {path}"
    detail = str(error)

    content = f"[red]{error_message}[/red]\n[yellow]{detail}[/yellow]"
    panel = Panel(
        content,
        title="[bold red]Error[/bold red]",
        border_style="red",
        expand=False,
    )
    console.print(panel)


def handle_pebble_api_error(
    console: Console, operation: str, error: ops.pebble.APIError
) -> None:
    """Handle Pebble APIError with standardized formatting.

    Args:
        console: Rich console for output
        operation: Description of the operation that failed
        error: The APIError that occurred
    """
    error_message = f"API error during {operation}"
    detail = str(error)

    content = f"[red]{error_message}[/red]\n[yellow]{detail}[/yellow]"
    panel = Panel(
        content,
        title="[bold red]API Error[/bold red]",
        border_style="red",
        expand=False,
    )
    console.print(panel)


def handle_generic_pebble_error(
    console: Console, operation: str, error: Exception, context: str = ""
) -> None:
    """Handle generic Pebble-related errors with standardized formatting.

    Args:
        console: Rich console for output
        operation: Description of the operation that failed
        error: The exception that occurred
        context: Optional additional context information
    """
    error_message = f"Error during {operation}"
    if context:
        error_message += f" ({context})"

    detail = f"{error.__class__.__name__}: {error}"

    content = f"[red]{error_message}[/red]\n[yellow]{detail}[/yellow]"
    panel = Panel(
        content,
        title="[bold red]Error[/bold red]",
        border_style="red",
        expand=False,
    )
    console.print(panel)


def safe_pebble_operation(
    console: Console, operation: Callable[[], T], operation_name: str, context: str = ""
) -> T | None:
    """Safely execute a Pebble operation with standardized error handling.

    Args:
        console: Rich console for error output
        operation: Function to execute that may raise Pebble exceptions
        operation_name: Human-readable description of the operation
        context: Optional additional context for error messages

    Returns:
        Result of the operation if successful, None if it failed
    """
    try:
        return operation()
    except ops.pebble.PathError as e:
        handle_pebble_path_error(console, operation_name, context or "unknown path", e)
        return None
    except ops.pebble.APIError as e:
        handle_pebble_api_error(console, operation_name, e)
        return None
    except Exception as e:
        handle_generic_pebble_error(console, operation_name, e, context)
        return None


def show_file_not_found_error(console: Console, path: str, command: str = "") -> None:
    """Show a standardized file not found error message.

    Args:
        console: Rich console for output
        path: Path that was not found
        command: Optional command name for context
    """
    prefix = f"{command}: " if command else ""
    error_message = f"{prefix}No such file or directory: {path}"

    console.print(f"[red]{error_message}[/red]")


def show_permission_error(
    console: Console, path: str, operation: str, command: str = ""
) -> None:
    """Show a standardized permission error message.

    Args:
        console: Rich console for output
        path: Path that caused the permission error
        operation: Operation that was attempted (e.g., "read", "write")
        command: Optional command name for context
    """
    prefix = f"{command}: " if command else ""
    error_message = f"{prefix}Permission denied: cannot {operation} {path}"

    console.print(f"[red]{error_message}[/red]")


def validate_file_exists(
    console: Console, client: Any, path: str, command: str = ""
) -> bool:
    """Validate that a file exists with standardized error handling.

    Args:
        console: Rich console for error output
        client: Pebble client instance
        path: Path to check
        command: Optional command name for context

    Returns:
        True if file exists, False otherwise (with error message displayed)
    """
    try:
        client.list_files(path, itself=True)
        return True
    except ops.pebble.PathError:
        show_file_not_found_error(console, path, command)
        return False
    except ops.pebble.APIError as e:
        handle_pebble_api_error(console, f"check file existence for {path}", e)
        return False


def show_usage_error(
    console: Console, command: str, usage: str, message: str = ""
) -> int:
    """Show a standardized usage error with command usage information.

    Args:
        console: Rich console for output
        command: Command name
        usage: Usage string for the command
        message: Optional additional error message

    Returns:
        Exit code 1
    """
    if message:
        console.print(f"[red]{command}: {message}[/red]")

    console.print(f"Usage: {usage}")
    return 1


def with_error_context(operation_name: str, context: str = ""):
    """Decorator to add standardized error handling to command methods.

    Args:
        operation_name: Human-readable name of the operation
        context: Optional additional context information

    Returns:
        Decorator function
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T | None]:
        def wrapper(self, *args, **kwargs) -> T | None:
            try:
                return func(self, *args, **kwargs)
            except ops.pebble.PathError as e:
                handle_pebble_path_error(
                    self.console, operation_name, context or "unknown path", e
                )
                return None
            except ops.pebble.APIError as e:
                handle_pebble_api_error(self.console, operation_name, e)
                return None
            except Exception as e:
                handle_generic_pebble_error(self.console, operation_name, e, context)
                return None

        return wrapper

    return decorator



"""Pebble CLI commands for Cascade."""

from .add import AddCommand
from .changes import ChangesCommand
from .check import CheckCommand
from .checks import ChecksCommand
from .health import HealthCommand
from .logs import LogsCommand
from .notice import NoticeCommand
from .notices import NoticesCommand
from .notify import NotifyCommand
from .pebble import PebbleCommand
from .plan import PlanCommand
from .pull import PullCommand
from .push import PushCommand
from .replan import ReplanCommand
from .restart import RestartCommand
from .services import ServicesCommand
from .signal import SignalCommand
from .start import StartCommand
from .startchecks import StartChecksCommand
from .stop import StopCommand
from .stopchecks import StopChecksCommand
from .tasks import TasksCommand

__all__ = [
    "AddCommand",
    "ChangesCommand",
    "CheckCommand",
    "ChecksCommand",
    "HealthCommand",
    "LogsCommand",
    "NoticeCommand",
    "NoticesCommand",
    "NotifyCommand",
    "PebbleCommand",
    "PlanCommand",
    "PullCommand",
    "PushCommand",
    "ReplanCommand",
    "RestartCommand",
    "ServicesCommand",
    "SignalCommand",
    "StartChecksCommand",
    "StartCommand",
    "StopChecksCommand",
    "StopCommand",
    "TasksCommand",
]



"""Filesystem write operations commands."""

from .copy import CopyCommand
from .make_dir import MakeDirCommand
from .move import MoveCommand
from .remove import RemoveCommand
from .remove_dir import RemoveDirCommand
from .touch import TouchCommand

__all__ = [
    "CopyCommand",
    "MakeDirCommand",
    "MoveCommand",
    "RemoveCommand",
    "RemoveDirCommand",
    "TouchCommand",
]



"""Integration tests for the SedCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a SedCommand instance."""
    yield pebble_shell.commands.SedCommand(shell=shell)


def test_name(command: pebble_shell.commands.SedCommand):
    assert command.name == "sed"


def test_category(command: pebble_shell.commands.SedCommand):
    assert command.category == "Text Processing"


def test_help(command: pebble_shell.commands.SedCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "sed" in output
    assert "Stream editor" in output
    assert "substitute" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "sed" in output
    assert "Stream editor" in output


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # sed with no args should fail and show usage
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should fail with usage message
    assert result == 1
    output = capture.get()
    assert any(
        msg in output for msg in ["usage", "Usage", "sed", "expression", "script"]
    )


def test_execute_simple_substitution_with_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test simple substitution with file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/root/admin/", "/etc/passwd"])

    # Should either succeed with substitution or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show modified content
        assert len(output.strip()) > 0
        # Should contain substituted text if root was present
        if "admin" in output:
            assert "admin" in output
    else:
        # Should fail if file doesn't exist or access denied
        assert result == 1
        output = capture.get()
        assert any(
            msg in output
            for msg in ["No such file", "cannot read", "error", "permission"]
        )


def test_execute_global_substitution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test global substitution
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/a/A/g", "/etc/hosts"])

    # Should either succeed with global substitution or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with all 'a' replaced with 'A'
        assert len(output.strip()) >= 0
    else:
        # Should fail if file doesn't exist
        assert result == 1


def test_execute_delete_lines(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test deleting lines
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/^#/d", "/etc/hosts"])

    # Should either succeed deleting comment lines or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content without comment lines
        assert len(output.strip()) >= 0
        # Should not contain lines starting with #
        lines = output.strip().split("\n") if output.strip() else []
        for line in lines:
            if line.strip():
                assert not line.strip().startswith("#")
    else:
        assert result == 1


def test_execute_print_specific_lines(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test printing specific lines
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "1,3p", "/etc/hosts"])

    # Should either succeed printing lines 1-3 or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show only lines 1-3
        assert len(output.strip()) >= 0
        # Should have limited output
        lines = output.strip().split("\n") if output.strip() else []
        assert len(lines) <= 3
    else:
        assert result == 1


def test_execute_address_range_substitution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test substitution with address range
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["1,5s/localhost/localserver/", "/etc/hosts"]
        )

    # Should either succeed with range substitution or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with substitution in first 5 lines
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_pattern_matching_substitution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test pattern matching substitution
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/localhost/s/127/192/", "/etc/hosts"]
        )

    # Should either succeed with pattern-based substitution or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with conditional substitution
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_multiple_commands(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test multiple sed commands
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=["-e", "s/root/admin/", "-e", "s/bin/usr/", "/etc/passwd"],
        )

    # Should either succeed with multiple commands or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with both substitutions applied
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_in_place_editing(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test in-place editing (dangerous, so we use a safe approach)
    temp_dir = tempfile.mkdtemp()
    nonexistent_path = temp_dir + "/nonexistent"
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["-i.bak", "s/test/TEST/", nonexistent_path]
        )

    # Should fail gracefully for nonexistent file
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["No such file", "cannot", "error"])


def test_execute_append_text(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test appending text
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["1a\\", "Added line", "/etc/hosts"]
        )

    # Should either succeed appending text or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with appended text
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_insert_text(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test inserting text
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["1i\\", "Inserted line", "/etc/hosts"]
        )

    # Should either succeed inserting text or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with inserted text
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_change_lines(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test changing lines
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["1c\\", "Changed line", "/etc/hosts"]
        )

    # Should either succeed changing lines or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with changed line
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_line_numbering(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test line numbering
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["=", "/etc/hosts"])

    # Should either succeed with line numbers or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with line numbers
        assert len(output.strip()) >= 0
        # May contain numeric line numbers
        if output.strip():
            lines = output.strip().split("\n")
            # Some lines should be numeric (line numbers)
            numeric_lines = [line for line in lines if line.strip().isdigit()]
            assert len(numeric_lines) >= 0
    else:
        assert result == 1


def test_execute_transform_characters(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test character transformation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["y/abc/ABC/", "/etc/hosts"])

    # Should either succeed with character transformation or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with character transformation
        assert len(output.strip()) >= 0
        # Should transform lowercase to uppercase
        if "a" in output or "b" in output or "c" in output:
            # Check if transformation occurred
            assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_hold_space_operations(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test hold space operations
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["1h;2g", "/etc/hosts"])

    # Should either succeed with hold space operations or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with hold space manipulation
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_next_line_operations(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test next line operations
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["n;s/test/TEST/", "/etc/hosts"])

    # Should either succeed with next line operations or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with next line processing
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_branching(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test branching operations
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=[":label;s/a/A/;t label", "/etc/hosts"]
        )

    # Should either succeed with branching or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with branching logic applied
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_quiet_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test quiet mode
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["-n", "/localhost/p", "/etc/hosts"]
        )

    # Should either succeed in quiet mode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show only explicitly printed lines
        assert len(output.strip()) >= 0
        # Should contain only lines matching localhost pattern
        lines = output.strip().split("\n") if output.strip() else []
        for line in lines:
            if line.strip():
                # Lines should contain localhost or be empty
                assert "localhost" in line.lower() or line.strip() == ""
    else:
        assert result == 1


def test_execute_extended_regex(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test extended regular expressions
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["-E", "s/[0-9]+/NUMBER/g", "/etc/hosts"]
        )

    # Should either succeed with extended regex or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with numbers replaced
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_script_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test script file execution
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["-f", "/nonexistent/script.sed", "/etc/hosts"]
        )

    # Should fail gracefully for nonexistent script file
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["No such file", "cannot", "error"])


def test_execute_invalid_regex(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test invalid regular expression
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/[/invalid/", "/etc/hosts"])

    # Should fail with invalid regex error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["Invalid", "error", "regex", "expression"])


def test_execute_invalid_command(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test invalid sed command
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["z/invalid/", "/etc/hosts"])

    # Should fail with invalid command error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "error", "command"])


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test with nonexistent file
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["s/test/TEST/", "/nonexistent/file.txt"]
        )

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["No such file", "cannot", "error"])


def test_execute_permission_denied_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test with permission denied file
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["s/test/TEST/", "/root/.ssh/id_rsa"]
        )

    # Should either succeed or fail with permission error
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "cannot", "error"])
    else:
        # May succeed if file is readable
        assert result == 0


def test_execute_empty_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test with empty file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/test/TEST/", "/dev/null"])

    # Should succeed with empty output
    assert result == 0
    output = capture.get()
    # Should produce no output for empty file
    assert len(output.strip()) == 0


def test_execute_binary_file_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test with binary file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/ELF/elf/", "/bin/ls"])

    # Should either succeed or handle binary gracefully
    if result == 0:
        output = capture.get()
        # Should process binary file (may contain binary data)
        assert len(output) >= 0
    else:
        # May fail for binary files
        assert result == 1


def test_execute_large_file_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test with large file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/usr/opt/", "/usr/bin/bash"])

    # Should either succeed processing large file or fail gracefully
    if result == 0:
        output = capture.get()
        # Should process large file efficiently
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_output_format_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test output format validation
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["s/localhost/localserver/", "/etc/hosts"]
        )

    # Should produce properly formatted output
    if result == 0:
        output = capture.get()
        # Should have valid text output
        assert len(output.strip()) >= 0
        if len(output.strip()) > 0:
            # Should be valid text format
            lines = output.strip().split("\n")
            assert len(lines) >= 1
    else:
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/a/A/g", "/etc/passwd"])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 100000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_concurrent_execution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test concurrent execution safety
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/test/TEST/", "/etc/hosts"])

    # Should handle concurrent execution safely
    if result == 0:
        output = capture.get()
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/[/invalid/", "/etc/hosts"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["Invalid", "error", "expression", "regex"])


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test signal handling during processing
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/test/TEST/", "/etc/hosts"])

    # Should handle signals appropriately
    if result == 0:
        output = capture.get()
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_locale_awareness(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test locale awareness
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/café/coffee/", "/etc/hosts"])

    # Should handle locale-specific characters
    if result == 0:
        output = capture.get()
        # Should process Unicode characters appropriately
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_line_ending_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test line ending handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/$/\\r/", "/etc/hosts"])

    # Should handle different line endings
    if result == 0:
        output = capture.get()
        # Should process line endings correctly
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_pattern_space_operations(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test pattern space operations
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["N;s/\\n/ /", "/etc/hosts"])

    # Should handle pattern space manipulation
    if result == 0:
        output = capture.get()
        # Should process pattern space correctly
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_address_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test address validation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["0s/test/TEST/", "/etc/hosts"])

    # Should validate addresses properly
    assert result == 1
    output = capture.get()
    # Should show address error
    assert any(msg in output for msg in ["invalid", "address", "error"])


def test_execute_escape_sequence_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test escape sequence handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/test/\\t\\n/", "/etc/hosts"])

    # Should handle escape sequences
    if result == 0:
        output = capture.get()
        # Should process escape sequences correctly
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_backreference_substitution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test backreference substitution
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=["s/\\(.*\\)localhost\\(.*\\)/\\1localserver\\2/", "/etc/hosts"],
        )

    # Should handle backreferences
    if result == 0:
        output = capture.get()
        # Should process backreferences correctly
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_command_line_parsing(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test command line parsing
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["s/test/TEST/", "/etc/hosts", "/etc/passwd"]
        )

    # Should handle multiple input files
    if result == 0:
        output = capture.get()
        # Should process multiple files
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_standard_input_processing(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test standard input processing
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/test/TEST/"])

    # Should handle standard input processing
    if result == 0:
        output = capture.get()
        # Should wait for input or process empty input
        assert len(output.strip()) >= 0
    else:
        # May fail without input
        assert result == 1


def test_execute_performance_optimization(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SedCommand,
):
    # Test performance optimization
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["s/a/A/g", "/etc/passwd"])

    # Should complete efficiently
    if result == 0:
        output = capture.get()
        # Should process efficiently
        assert len(output) >= 0
    else:
        assert result == 1



"""Implementation of EditCommand."""

from __future__ import annotations

import io
import os
import shutil
import subprocess
import tempfile
from typing import TYPE_CHECKING, Union

import ops
from rich.panel import Panel

from ...utils.command_helpers import handle_help_flag
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class EditCommand(Command):
    """Command for editing remote files locally with automatic sync."""

    name = "edit"
    help = "Edit a remote file locally and push changes back. Usage: edit REMOTE_PATH"
    category = "Built-in Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute the edit command to edit a remote file locally."""
        if handle_help_flag(self, args):
            return 0
        console = self.console
        if not args:
            console.print(
                Panel(
                    "No remote file specified.",
                    title="[b red]edit Error[/b red]",
                    style="red",
                )
            )
            return 1
        remote_path = args[0]
        cwd = self.shell.current_directory
        if not os.path.isabs(remote_path):
            remote_path = os.path.normpath(os.path.join(cwd, remote_path))
        with tempfile.NamedTemporaryFile(delete=False, mode="wb") as tmp:
            tmp_path = tmp.name
            with client.pull(remote_path, encoding=None) as remote_f:
                shutil.copyfileobj(remote_f, tmp)
        editor = os.environ.get("EDITOR", "pico")
        before = os.stat(tmp_path).st_mtime
        subprocess.run([editor, tmp_path], check=True)  # noqa: S603
        after = os.stat(tmp_path).st_mtime
        if after != before:
            with open(tmp_path, "rb") as f:
                client.push(remote_path, f)
            console.print(
                Panel(
                    f"[green]Saved changes to {remote_path}[/green]",
                    title="[b]edit[/b]",
                    style="green",
                )
            )
        else:
            console.print(
                Panel(
                    f"[yellow]No changes made to {remote_path}[/yellow]",
                    title="[b]edit[/b]",
                    style="yellow",
                )
            )
        return 0



"""Implementation of MountCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ...utils.command_helpers import handle_help_flag
from ...utils.proc_reader import parse_proc_mounts_file
from ...utils.table_builder import create_enhanced_table
from .._base import Command

if TYPE_CHECKING:
    import ops
    import shimmer


class MountCommand(Command):
    """Show mounted filesystems."""

    name = "mount"
    help = "Show mounted filesystems"
    category = "System"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute mount command."""
        if handle_help_flag(self, args):
            return 0
        try:
            mounts = parse_proc_mounts_file(client)

            table = create_enhanced_table()
            table.add_column("Device", style="cyan", no_wrap=True)
            table.add_column("Mount Point", style="green", no_wrap=True)
            table.add_column("Type", style="yellow", no_wrap=True)
            table.add_column("Options", style="blue", no_wrap=False)

            for mount_info in mounts:
                device = mount_info["device"]
                mount_point = mount_info["mountpoint"]
                fs_type = mount_info["fstype"]
                options = mount_info["options"]

                # Color code different filesystem types
                if fs_type in ["ext4", "xfs", "btrfs"]:
                    fs_type_colored = f"[green]{fs_type}[/green]"
                elif fs_type in ["proc", "sysfs", "tmpfs"]:
                    fs_type_colored = f"[blue]{fs_type}[/blue]"
                elif fs_type in ["devpts", "devtmpfs"]:
                    fs_type_colored = f"[yellow]{fs_type}[/yellow]"
                else:
                    fs_type_colored = fs_type

                table.add_row(device, mount_point, fs_type_colored, options)

            self.console.print(table.build())

        except Exception as e:
            self.console.print(f"Error reading mount information: {e}")
            return 1

        return 0



"""Integration tests for the PrintfCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a PrintfCommand instance."""
    yield pebble_shell.commands.PrintfCommand(shell=shell)


def test_name(command: pebble_shell.commands.PrintfCommand):
    assert command.name == "printf"


def test_category(command: pebble_shell.commands.PrintfCommand):
    assert command.category == "Built-in Commands"


def test_help(command: pebble_shell.commands.PrintfCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "Format and print data" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "Format and print data" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])
    # Should fail with missing format string error
    assert result == 1
    assert "missing format string" in capture.get()


def test_execute_simple_string(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Hello World"])
    # Should succeed and print the string
    assert result == 0
    assert capture.get() == "Hello World"


def test_execute_string_with_newline(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Hello\\nWorld"])
    # Should succeed and process newline escape
    assert result == 0
    assert capture.get() == "Hello\nWorld"


def test_execute_string_with_tab(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Hello\\tWorld"])
    # Should succeed and process tab escape
    assert result == 0
    assert capture.get() == "Hello\tWorld"


def test_execute_string_with_backslash(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Hello\\\\World"])
    # Should succeed and process backslash escape
    assert result == 0
    assert capture.get() == "Hello\\World"


def test_execute_string_substitution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Hello %s", "World"])
    # Should succeed and substitute string
    assert result == 0
    assert capture.get() == "Hello World"


def test_execute_decimal_integer(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Number: %d", "42"])
    # Should succeed and format decimal integer
    assert result == 0
    assert capture.get() == "Number: 42"


def test_execute_decimal_integer_invalid(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Number: %d", "invalid"])
    # Should succeed but use default value 0 for invalid input
    assert result == 0
    assert capture.get() == "Number: 0"


def test_execute_octal_integer(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Octal: %o", "8"])
    # Should succeed and format as octal
    assert result == 0
    assert capture.get() == "Octal: 10"


def test_execute_hex_lowercase(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Hex: %x", "255"])
    # Should succeed and format as lowercase hex
    assert result == 0
    assert capture.get() == "Hex: ff"


def test_execute_hex_uppercase(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Hex: %X", "255"])
    # Should succeed and format as uppercase hex
    assert result == 0
    assert capture.get() == "Hex: FF"


def test_execute_floating_point(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Float: %f", "3.14"])
    # Should succeed and format floating point
    assert result == 0
    assert capture.get() == "Float: 3.14"


def test_execute_floating_point_invalid(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Float: %f", "invalid"])
    # Should succeed but use default value 0.0 for invalid input
    assert result == 0
    assert capture.get() == "Float: 0.0"


def test_execute_character(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Char: %c", "Hello"])
    # Should succeed and use first character
    assert result == 0
    assert capture.get() == "Char: H"


def test_execute_literal_percent(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Percent: %%"])
    # Should succeed and print literal percent
    assert result == 0
    assert capture.get() == "Percent: %"


def test_execute_multiple_arguments(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["%d + %d = %d", "2", "3", "5"])
    # Should succeed and substitute all arguments
    assert result == 0
    assert capture.get() == "2 + 3 = 5"


def test_execute_insufficient_arguments(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["%s %d %f", "hello"])
    # Should succeed but use defaults for missing arguments
    assert result == 0
    assert capture.get() == "hello 0 0.0"


def test_execute_excess_arguments(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["%s", "hello", "world", "extra"])
    # Should succeed and ignore extra arguments
    assert result == 0
    assert capture.get() == "hello"


def test_execute_unknown_specifier(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Test: %z", "value"])
    # Should succeed and handle unknown specifier literally
    assert result == 0
    assert capture.get() == "Test: %z"


def test_execute_complex_format(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=[
                "User: %s, ID: %d, Hex: 0x%x, Float: %f%%\\n",
                "alice",
                "1001",
                "16",
                "98.6",
            ],
        )
    # Should succeed with complex formatting
    assert result == 0
    assert capture.get() == "User: alice, ID: 1001, Hex: 0x10, Float: 98.6%\n"


def test_execute_empty_format_string(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[""])
    # Should succeed with empty format string
    assert result == 0
    assert capture.get() == ""


def test_execute_only_escape_sequences(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["\\n\\t\\\\"])
    # Should succeed and process all escape sequences
    assert result == 0
    assert capture.get() == "\n\t\\"


def test_execute_character_empty_arg(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Char: %c", ""])
    # Should succeed but print nothing for empty character argument
    assert result == 0
    assert capture.get() == "Char: "


def test_execute_i_specifier(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PrintfCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["Integer: %i", "42"])
    # Should succeed and format with %i (same as %d)
    assert result == 0
    assert capture.get() == "Integer: 42"



"""Integration tests for the UnzipCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create an UnzipCommand instance."""
    yield pebble_shell.commands.UnzipCommand(shell=shell)


def test_name(command: pebble_shell.commands.UnzipCommand):
    assert command.name == "unzip"


def test_category(command: pebble_shell.commands.UnzipCommand):
    assert command.category == "Compression"


def test_help(command: pebble_shell.commands.UnzipCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "extract files from ZIP" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "unzip" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=[])
    assert result == 1
    # Should show error message about missing file operand


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["/nonexistent/file.zip"])
    # Should fail for nonexistent file
    assert result == 1


def test_execute_with_list_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-l", "/etc/passwd"])
    # Should fail since /etc/passwd is not a ZIP file
    assert result == 1


def test_execute_with_test_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-t", "/etc/passwd"])
    # Should fail since /etc/passwd is not a ZIP file
    assert result == 1


def test_execute_with_quiet_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-q", "/etc/passwd"])
    # Should fail since /etc/passwd is not a ZIP file
    assert result == 1


def test_execute_with_overwrite_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-o", "/etc/passwd"])
    # Should fail since /etc/passwd is not a ZIP file
    assert result == 1


def test_execute_with_directory_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    temp_dir = tempfile.mkdtemp()
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-d", temp_dir, "/etc/passwd"])
    # Should fail since /etc/passwd is not a ZIP file
    assert result == 1


def test_execute_with_combined_flags(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnzipCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-qo", "/etc/passwd"])
    # Should fail since /etc/passwd is not a ZIP file
    assert result == 1



"""Implementation of MktempCommand."""

from __future__ import annotations

import os
import uuid
from io import StringIO
from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags
from ...utils.theme import get_theme
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class MktempCommand(Command):
    """Implementation of mktemp command."""

    name = "mktemp"
    help = "Create temporary files or directories"
    category = "File Utilities"

    def show_help(self):
        """Show command help."""
        help_text = """Create temporary files or directories.

Usage: mktemp [OPTIONS] [TEMPLATE]

Description:
    Create a temporary file or directory and print its name.
    TEMPLATE may contain XXXXXX which will be replaced with random characters.

Options:
    -d, --directory Create a directory instead of a file
    -p DIR          Use DIR as prefix for temporary files
    -t              Treat TEMPLATE as a single file component relative to temp dir
    -h, --help      Show this help message

Examples:
    mktemp
    mktemp -d
    mktemp fileXXXXXX.txt
    mktemp -d -t tmpdir.XXXXXX
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the mktemp command."""
        if handle_help_flag(self, args):
            return 0

        parse_result = parse_flags(
            args,
            {
                "d": bool,  # create directory
                "directory": bool,
                "p": str,  # prefix directory
                "t": bool,  # template relative to temp dir
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        create_dir = flags.get("d", False) or flags.get("directory", False)
        prefix_dir = flags.get("p", None)
        use_temp_dir = flags.get("t", False)
        template = positional_args[0] if positional_args else None

        try:
            # Determine template and directory
            if use_temp_dir or not template:
                # Use system temp directory
                temp_base = "/tmp"  # noqa: S108
                template_name = template or "tmp.XXXXXX"
            else:
                # Use provided template path
                if "/" in template:
                    temp_base = os.path.dirname(template)
                    template_name = os.path.basename(template)
                else:
                    temp_base = "."
                    template_name = template

            if prefix_dir:
                temp_base = prefix_dir

            # Create the temporary file/directory on the remote system
            # Since we can't directly create temp files on remote, we'll generate a name

            if "XXXXXX" in template_name:
                # Replace XXXXXX with random characters
                random_part = str(uuid.uuid4())[:6]
                final_name = template_name.replace("XXXXXX", random_part)
            else:
                final_name = f"{template_name}.{str(uuid.uuid4())[:6]}"

            full_path = f"{temp_base.rstrip('/')}/{final_name}"

            if create_dir:
                # Create directory
                try:
                    client.make_dir(full_path, make_parents=True)
                    self.console.print(full_path)
                    return 0
                except Exception as e:
                    self.console.print(
                        get_theme().error_text(
                            f"mktemp: cannot create directory {full_path}: {e}"
                        )
                    )
                    return 1
            else:
                # Create file
                try:
                    # Create empty file by pushing empty content
                    empty_content = StringIO("")
                    client.push(full_path, empty_content)
                    self.console.print(full_path)
                    return 0
                except Exception as e:
                    self.console.print(
                        get_theme().error_text(
                            f"mktemp: cannot create file {full_path}: {e}"
                        )
                    )
                    return 1

        except Exception as e:
            self.console.print(get_theme().error_text(f"mktemp: {e}"))
            return 1



"""Real-time system monitoring dashboard for Cascade shell."""

from __future__ import annotations

import dataclasses
import datetime
import logging
import threading
import time
from typing import TYPE_CHECKING

from rich.align import Align
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

if TYPE_CHECKING:
    from pebble_shell.shell import PebbleShell


@dataclasses.dataclass
class SystemStats:
    """Container for system statistics."""

    # CPU stats
    cpu_user: float = 0.0
    cpu_system: float = 0.0
    cpu_idle: float = 0.0
    cpu_iowait: float = 0.0

    # Memory stats
    mem_total: int = 0
    mem_used: int = 0
    mem_free: int = 0
    mem_available: int = 0
    mem_buffers: int = 0
    mem_cached: int = 0

    # Load averages
    load_1min: float = 0.0
    load_5min: float = 0.0
    load_15min: float = 0.0

    # Process stats
    process_count: int = 0
    running_processes: int = 0

    # Disk stats
    disk_reads: int = 0
    disk_writes: int = 0
    disk_read_kb: int = 0
    disk_write_kb: int = 0

    # Network stats (basic)
    network_interfaces: dict[str, dict[str, int]] = dataclasses.field(
        default_factory=dict
    )

    # System info
    uptime_seconds: int = 0
    cpu_cores: int = 0


class SystemDashboard:
    """Real-time system monitoring dashboard."""

    def __init__(self, shell: PebbleShell):
        self.shell = shell
        self.stats = SystemStats()
        self.running = False
        self.update_thread = None
        self.update_interval = 1.0  # seconds

        # Historical data for trends
        self.cpu_history: list[float] = []
        self.memory_history: list[float] = []
        self.load_history: list[float] = []
        self.history_length = 20

        # Alert thresholds
        self.alerts = {
            "cpu_high": 80.0,
            "memory_high": 85.0,
            "load_high": 1.0,
            "disk_high": 90.0,
        }

    def start(self):
        """Start the real-time dashboard."""
        self.running = True
        self.update_thread = threading.Thread(target=self._update_loop, daemon=True)
        self.update_thread.start()

        try:
            with Live(
                self._create_dashboard(), refresh_per_second=2, screen=True
            ) as live:
                while self.running:
                    live.update(self._create_dashboard())
                    time.sleep(0.5)
        except KeyboardInterrupt:
            self.stop()

    def stop(self):
        """Stop the dashboard."""
        self.running = False
        if self.update_thread:
            self.update_thread.join(timeout=1.0)

    def _update_loop(self):
        """Background thread for updating system statistics."""
        while self.running:
            try:
                self._update_stats()
                time.sleep(self.update_interval)
            except Exception as e:  # noqa: PERF203
                # Log errors but keep dashboard running
                logging.warning("Dashboard update failed: %s", e)

    def _update_stats(self):
        """Update system statistics from remote system."""
        try:
            # Get CPU stats
            self._update_cpu_stats()

            # Get memory stats
            self._update_memory_stats()

            # Get load averages
            self._update_load_stats()

            # Get process stats
            self._update_process_stats()

            # Get disk stats
            self._update_disk_stats()

            # Get network stats
            self._update_network_stats()

            # Get system info
            self._update_system_info()

            # Update history
            self._update_history()

        except Exception as e:
            # Log errors but don't crash dashboard
            logging.warning("Failed to update dashboard stats: %s", e)

    def _update_cpu_stats(self):
        """Update CPU statistics."""
        with self.shell.client.pull("/proc/stat") as file:
            content = file.read()
        assert isinstance(content, str)
        for line in content.splitlines():
            if line.startswith("cpu "):
                parts = line.split()
                if len(parts) >= 5:
                    total = sum(int(parts[i]) for i in range(1, 5))
                    if total > 0:
                        self.stats.cpu_user = (int(parts[1]) / total) * 100
                        self.stats.cpu_system = (int(parts[3]) / total) * 100
                        self.stats.cpu_idle = (int(parts[4]) / total) * 100
                        self.stats.cpu_iowait = (
                            (int(parts[5]) / total) * 100 if len(parts) > 5 else 0
                        )
                break

    def _update_memory_stats(self):
        """Update memory statistics."""
        with self.shell.client.pull("/proc/meminfo") as file:
            content = file.read()
        assert isinstance(content, str)

        mem_stats: dict[str, int] = {}
        for line in content.splitlines():
            if ":" in line:
                key, value = line.split(":", 1)
                key = key.strip()
                value = value.strip().split()[0]
                try:
                    mem_stats[key] = int(value)
                except ValueError:
                    continue

        self.stats.mem_total = mem_stats.get("MemTotal", 0)
        self.stats.mem_free = mem_stats.get("MemFree", 0)
        self.stats.mem_available = mem_stats.get("MemAvailable", self.stats.mem_free)
        self.stats.mem_buffers = mem_stats.get("Buffers", 0)
        self.stats.mem_cached = mem_stats.get("Cached", 0)
        self.stats.mem_used = (
            self.stats.mem_total - self.stats.mem_available
            if self.stats.mem_total > 0
            else 0
        )

    def _update_load_stats(self):
        """Update load average statistics."""
        with self.shell.client.pull("/proc/loadavg") as file:
            content = file.read().strip()
        assert isinstance(content, str)

        parts = content.split()
        if len(parts) >= 3:
            self.stats.load_1min = float(parts[0])
            self.stats.load_5min = float(parts[1])
            self.stats.load_15min = float(parts[2])

            # Get process info
            if len(parts) >= 5:
                process_info = parts[3].split("/")
                if len(process_info) == 2:
                    self.stats.running_processes = int(process_info[0])
                    self.stats.process_count = int(process_info[1])

    def _update_process_stats(self):
        """Update process statistics."""
        proc_entries = self.shell.client.list_files("/proc")
        self.stats.process_count = sum(
            1 for entry in proc_entries if entry.name.isdigit()
        )

    def _update_disk_stats(self):
        """Update disk I/O statistics."""
        with self.shell.client.pull("/proc/diskstats") as file:
            content = file.read()

        total_reads = 0
        total_writes = 0
        total_read_sectors = 0
        total_write_sectors = 0

        for line in content.splitlines():
            if line.strip():
                parts = line.split()
                if len(parts) >= 14:
                    try:
                        total_reads += int(parts[3])
                        total_writes += int(parts[7])
                        total_read_sectors += int(parts[5])
                        total_write_sectors += int(parts[9])
                    except (ValueError, IndexError):
                        continue

        self.stats.disk_reads = total_reads
        self.stats.disk_writes = total_writes
        self.stats.disk_read_kb = int(total_read_sectors * 0.5)  # Convert sectors to KB
        self.stats.disk_write_kb = int(total_write_sectors * 0.5)

    def _update_network_stats(self):
        """Update network interface statistics."""
        with self.shell.client.pull("/proc/net/dev") as file:
            content = file.read()
        assert isinstance(content, str)

        self.stats.network_interfaces = {}
        for line in content.splitlines()[2:]:  # Skip header lines
            if ":" in line:
                parts = line.split(":")
                if len(parts) >= 2:
                    interface = parts[0].strip()
                    stats = parts[1].split()
                    if len(stats) >= 16:
                        try:
                            self.stats.network_interfaces[interface] = {
                                "rx_bytes": int(stats[0]),
                                "rx_packets": int(stats[1]),
                                "tx_bytes": int(stats[8]),
                                "tx_packets": int(stats[9]),
                            }
                        except (ValueError, IndexError):
                            continue

    def _update_system_info(self):
        """Update system information."""
        with self.shell.client.pull("/proc/cpuinfo") as file:
            content = file.read()
        assert isinstance(content, str)

        self.stats.cpu_cores = sum(
            1 for line in content.splitlines() if line.startswith("processor")
        )

        with self.shell.client.pull("/proc/uptime") as file:
            content = file.read().strip()

        if content:
            self.stats.uptime_seconds = int(float(content.split()[0]))

    def _update_history(self):
        """Update historical data for trends."""
        cpu_usage = self.stats.cpu_user + self.stats.cpu_system
        memory_usage = (
            (self.stats.mem_used / self.stats.mem_total * 100)
            if self.stats.mem_total > 0
            else 0
        )

        self.cpu_history.append(cpu_usage)
        self.memory_history.append(memory_usage)
        self.load_history.append(self.stats.load_1min)

        # Keep history length
        if len(self.cpu_history) > self.history_length:
            self.cpu_history.pop(0)
            self.memory_history.pop(0)
            self.load_history.pop(0)

    def _create_dashboard(self) -> Layout:
        """Create the dashboard layout."""
        layout = Layout()

        layout.split_column(
            Layout(name="header", size=3),
            Layout(name="main"),
            Layout(name="footer", size=3),
        )

        layout["main"].split_row(
            Layout(name="left", ratio=1), Layout(name="right", ratio=1)
        )

        layout["left"].split_column(Layout(name="cpu_memory"), Layout(name="processes"))

        layout["right"].split_column(Layout(name="load_disk"), Layout(name="network"))

        layout["header"].update(self._create_header())
        layout["cpu_memory"].update(self._create_cpu_memory_panel())
        layout["processes"].update(self._create_processes_panel())
        layout["load_disk"].update(self._create_load_disk_panel())
        layout["network"].update(self._create_network_panel())
        layout["footer"].update(self._create_footer())

        return layout

    def _create_header(self) -> Panel:
        """Create the header panel."""
        uptime_str = self._format_uptime(self.stats.uptime_seconds)

        header_text = Text()
        header_text.append("🖥️  Cascade System Dashboard", style="bold cyan")
        header_text.append(" | ")
        header_text.append(f"Uptime: {uptime_str}", style="yellow")
        header_text.append(" | ")
        header_text.append(f"CPU Cores: {self.stats.cpu_cores}", style="green")
        header_text.append(" | ")
        header_text.append(
            f"Last Update: {datetime.datetime.now().strftime('%H:%M:%S')}", style="dim"
        )

        return Panel(
            Align.center(header_text), style="bold blue", border_style="bright_blue"
        )

    def _create_cpu_memory_panel(self) -> Panel:
        """Create CPU and Memory panel."""
        table = Table(show_header=False, box=None, padding=(0, 1))
        table.add_column("Metric", style="cyan", no_wrap=True)
        table.add_column("Value", style="yellow", no_wrap=True)
        table.add_column("Bar", style="green", no_wrap=True)

        cpu_total = self.stats.cpu_user + self.stats.cpu_system
        cpu_bar = self._create_progress_bar(cpu_total, 100, 20)
        cpu_style = "red" if cpu_total > self.alerts["cpu_high"] else "green"
        table.add_row(
            "CPU Usage", f"{cpu_total:.1f}%", f"[{cpu_style}]{cpu_bar}[/{cpu_style}]"
        )

        mem_usage = (
            (self.stats.mem_used / self.stats.mem_total * 100)
            if self.stats.mem_total > 0
            else 0
        )
        mem_bar = self._create_progress_bar(mem_usage, 100, 20)
        mem_style = "red" if mem_usage > self.alerts["memory_high"] else "green"
        table.add_row(
            "Memory", f"{mem_usage:.1f}%", f"[{mem_style}]{mem_bar}[/{mem_style}]"
        )

        mem_total_mb = self.stats.mem_total // 1024
        mem_used_mb = self.stats.mem_used // 1024
        table.add_row("Memory Details", f"{mem_used_mb}M / {mem_total_mb}M", "")

        return Panel(table, title="💻 CPU & Memory", border_style="bright_blue")

    def _create_processes_panel(self) -> Panel:
        """Create processes panel."""
        table = Table(show_header=False, box=None, padding=(0, 1))
        table.add_column("Metric", style="cyan", no_wrap=True)
        table.add_column("Value", style="yellow", no_wrap=True)

        table.add_row("Total Processes", str(self.stats.process_count))
        table.add_row("Running Processes", str(self.stats.running_processes))

        if self.stats.cpu_cores > 0:
            load_per_core = self.stats.load_1min / self.stats.cpu_cores
            load_style = "red" if load_per_core > self.alerts["load_high"] else "green"
            table.add_row(
                "Load per Core", f"[{load_style}]{load_per_core:.2f}[/{load_style}]"
            )
        else:
            table.add_row("Load per Core", "N/A")

        return Panel(table, title="📊 Processes", border_style="bright_blue")

    def _create_load_disk_panel(self) -> Panel:
        """Create load averages and disk I/O panel."""
        table = Table(show_header=False, box=None, padding=(0, 1))
        table.add_column("Metric", style="cyan", no_wrap=True)
        table.add_column("Value", style="yellow", no_wrap=True)

        table.add_row("Load 1min", f"{self.stats.load_1min:.2f}")
        table.add_row("Load 5min", f"{self.stats.load_5min:.2f}")
        table.add_row("Load 15min", f"{self.stats.load_15min:.2f}")

        table.add_row("Disk Reads", f"{self.stats.disk_reads:,}")
        table.add_row("Disk Writes", f"{self.stats.disk_writes:,}")
        table.add_row("Disk Read KB", f"{self.stats.disk_read_kb:,}")
        table.add_row("Disk Write KB", f"{self.stats.disk_write_kb:,}")

        return Panel(table, title="⚡ Load & Disk", border_style="bright_blue")

    def _create_network_panel(self) -> Panel:
        """Create network interfaces panel."""
        table = Table(show_header=False, box=None, padding=(0, 1))
        table.add_column("Interface", style="cyan", no_wrap=True)
        table.add_column("RX/TX", style="yellow", no_wrap=True)

        for interface, stats in list(self.stats.network_interfaces.items())[
            :5
        ]:  # Show top 5
            rx_mb = stats["rx_bytes"] // (1024 * 1024)
            tx_mb = stats["tx_bytes"] // (1024 * 1024)
            table.add_row(interface, f"{rx_mb}M / {tx_mb}M")

        if not self.stats.network_interfaces:
            table.add_row("No data", "N/A")

        return Panel(table, title="🌐 Network", border_style="bright_blue")

    def _create_footer(self) -> Panel:
        """Create the footer panel with alerts and controls."""
        footer_text = Text()
        footer_text.append("Press ", style="dim")
        footer_text.append("Ctrl+C", style="bold red")
        footer_text.append(" to exit dashboard | ", style="dim")
        footer_text.append("Refresh: 2s", style="green")

        alerts: list[str] = []
        if self.stats.cpu_user + self.stats.cpu_system > self.alerts["cpu_high"]:
            alerts.append("High CPU")
        if (
            self.stats.mem_total > 0
            and (self.stats.mem_used / self.stats.mem_total * 100)
            > self.alerts["memory_high"]
        ):
            alerts.append("High Memory")
        if self.stats.load_1min > self.alerts["load_high"]:
            alerts.append("High Load")

        if alerts:
            footer_text.append(" | Alerts: ", style="dim")
            footer_text.append(", ".join(alerts), style="bold red")

        return Panel(Align.center(footer_text), style="dim", border_style="bright_blue")

    def _create_progress_bar(self, value: float, max_value: float, width: int) -> str:
        """Create a text-based progress bar."""
        if max_value <= 0:
            return "█" * width

        filled = int((value / max_value) * width)
        bar = "█" * filled + "░" * (width - filled)
        return bar

    def _format_uptime(self, seconds: int) -> str:
        """Format uptime in human readable format."""
        days = seconds // 86400
        hours = (seconds % 86400) // 3600
        minutes = (seconds % 3600) // 60

        if days > 0:
            return f"{days}d {hours}h {minutes}m"
        if hours > 0:
            return f"{hours}h {minutes}m"
        return f"{minutes}m"



"""Implementation of AddgroupCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags, safe_read_file
from .._base import Command

if TYPE_CHECKING:
    import shimmer


# Standard Unix shadow file values
LOCKED_ACCOUNT_HASH = "!"  # Standard indicator for locked/disabled account


# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class AddgroupCommand(Command):
    """Implementation of addgroup command."""

    name = "addgroup"
    help = "Add a new group to the system"
    category = "User Management"

    def show_help(self):
        """Show command help."""
        help_text = """Add a new group to the system.

Usage: addgroup [OPTIONS] GROUP
       addgroup [OPTIONS] USER GROUP

Description:
    Add a new group to the system by modifying /etc/group and /etc/gshadow.
    Can also add an existing user to an existing group.

Options:
    --gid GID           Specify group ID (default: auto-assign)
    --system            Create a system group (GID < 1000)
    --quiet             Suppress informational messages
    -h, --help          Show this help message

Examples:
    addgroup mygroup            # Create new group
    addgroup --gid 1001 staff   # Create group with specific GID
    addgroup --system daemon    # Create system group
    addgroup user wheel         # Add user to existing group
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the addgroup command."""
        if handle_help_flag(self, args):
            return 0

        if not args:
            self.console.print("[red]addgroup: missing group name[/red]")
            return 1

        parse_result = parse_flags(
            args,
            {
                "gid": int,  # group ID
                "system": bool,  # system group
                "quiet": bool,  # quiet mode
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        if len(positional_args) == 1:
            # Creating new group
            group_name = positional_args[0]
            user_name = None
        elif len(positional_args) == 2:
            # Adding user to existing group
            user_name = positional_args[0]
            group_name = positional_args[1]
        else:
            self.console.print("[red]addgroup: invalid arguments[/red]")
            return 1

        gid = flags.get("gid")
        system_group = flags.get("system", False)
        quiet = flags.get("quiet", False)

        try:
            if user_name:
                # Add user to existing group
                return self._add_user_to_group(client, user_name, group_name, quiet)
            else:
                # Create new group
                return self._create_group(client, group_name, gid, system_group, quiet)

        except Exception as e:
            self.console.print(f"[red]addgroup: {e}[/red]")
            return 1

    def _create_group(
        self,
        client: ClientType,
        group_name: str,
        gid: int | None,
        system_group: bool,
        quiet: bool,
    ) -> int:
        """Create a new group."""
        # Read existing /etc/group
        group_content = safe_read_file(client, "/etc/group")
        if group_content is None:
            self.console.print("[red]addgroup: cannot read /etc/group[/red]")
            return 1

        # Check if group already exists
        for line in group_content.strip().split("\n"):
            if line and line.split(":")[0] == group_name:
                self.console.print(
                    f"[red]addgroup: group '{group_name}' already exists[/red]"
                )
                return 1

        # Determine GID
        if gid is None:
            gid = self._find_available_gid(group_content, system_group)

        # Check if GID is already in use
        for line in group_content.strip().split("\n"):
            if line and len(line.split(":")) >= 3:
                existing_gid = line.split(":")[2]
                if existing_gid == str(gid):
                    self.console.print(f"[red]addgroup: GID {gid} already in use[/red]")
                    return 1

        # Create group entry
        group_entry = f"{group_name}:x:{gid}:"

        # Read existing /etc/gshadow
        gshadow_content = safe_read_file(client, "/etc/gshadow")
        if gshadow_content is None:
            # Create minimal gshadow if it doesn't exist
            gshadow_content = ""

        # Create gshadow entry
        gshadow_entry = f"{group_name}:!::"

        # Write files (in Cascade, we simulate this)
        if not quiet:
            self.console.print(
                f"[green]Adding group '{group_name}' (GID {gid})[/green]"
            )

        self.console.print("[yellow]addgroup: would write to /etc/group:[/yellow]")
        self.console.print(f"[dim]{group_entry}[/dim]")

        self.console.print("[yellow]addgroup: would write to /etc/gshadow:[/yellow]")
        self.console.print(f"[dim]{gshadow_entry}[/dim]")

        if not quiet:
            self.console.print(
                f"[green]Group '{group_name}' created successfully[/green]"
            )

        return 0

    def _add_user_to_group(
        self, client: ClientType, user_name: str, group_name: str, quiet: bool
    ) -> int:
        """Add existing user to existing group."""
        # Read /etc/passwd to verify user exists
        passwd_content = safe_read_file(client, "/etc/passwd")
        if passwd_content is None:
            self.console.print("[red]addgroup: cannot read /etc/passwd[/red]")
            return 1

        user_exists = False
        for line in passwd_content.strip().split("\n"):
            if line and line.split(":")[0] == user_name:
                user_exists = True
                break

        if not user_exists:
            self.console.print(
                f"[red]addgroup: user '{user_name}' does not exist[/red]"
            )
            return 1

        # Read /etc/group to verify group exists and add user
        group_content = safe_read_file(client, "/etc/group")
        if group_content is None:
            self.console.print("[red]addgroup: cannot read /etc/group[/red]")
            return 1

        new_lines = []
        group_found = False

        for line in group_content.strip().split("\n"):
            if not line:
                continue

            parts = line.split(":")
            if len(parts) >= 4 and parts[0] == group_name:
                group_found = True
                # Check if user is already in group
                members = parts[3].split(",") if parts[3] else []
                if user_name in members:
                    self.console.print(
                        f"[yellow]addgroup: user '{user_name}' is already "
                        f"in group '{group_name}'[/yellow]"
                    )
                    return 0

                # Add user to group
                if members and members[0]:  # Group has existing members
                    new_members = ",".join([*members, user_name])
                else:  # Group has no members
                    new_members = user_name

                new_line = f"{parts[0]}:{parts[1]}:{parts[2]}:{new_members}"
                new_lines.append(new_line)

                if not quiet:
                    self.console.print(
                        f"[green]Adding user '{user_name}' to group '{group_name}'[/green]"
                    )

                self.console.print(
                    "[yellow]addgroup: would update /etc/group line:[/yellow]"
                )
                self.console.print(f"[dim]{new_line}[/dim]")
            else:
                new_lines.append(line)

        if not group_found:
            self.console.print(
                f"[red]addgroup: group '{group_name}' does not exist[/red]"
            )
            return 1

        if not quiet:
            self.console.print(
                f"[green]User '{user_name}' added to group '{group_name}' "
                f"successfully[/green]"
            )

        return 0

    def _find_available_gid(self, group_content: str, system_group: bool) -> int:
        """Find next available GID."""
        used_gids = set()

        for line in group_content.strip().split("\n"):
            if line and len(line.split(":")) >= 3:
                try:
                    gid = int(line.split(":")[2])
                    used_gids.add(gid)
                except ValueError:
                    continue

        # System groups: 100-999, regular groups: 1000+
        start_gid = 100 if system_group else 1000
        max_gid = 999 if system_group else 65534

        for gid in range(start_gid, max_gid + 1):
            if gid not in used_gids:
                return gid

        raise Exception(f"No available GID in range {start_gid}-{max_gid}")



"""Implementation of run-parts command."""

from __future__ import annotations

import string
from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags
from .._base import Command

if TYPE_CHECKING:
    import shimmer

    from pebble_shell.shell import PebbleShell

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class RunPartsCommand(Command):
    """Implementation of run-parts command."""

    name = "run-parts"
    help = "Run all executable files in a directory"
    category = "Process Management"

    def __init__(self, shell: PebbleShell) -> None:
        super().__init__(shell)

    def show_help(self):
        """Show command help."""
        help_text = """Run all executable files in a directory.

Usage: run-parts [OPTIONS] DIRECTORY

Description:
    Run all executable files in DIRECTORY in lexicographic order.
    Commonly used for running scripts in /etc/cron.daily, etc.

Options:
    --test          Print names of files that would be run, but don't run them
    --list          Print names of files that would be run (same as --test)
    -v, --verbose   Print name of each file before running it
    --report        Print name of each file as it is run
    --exit-on-error Exit if any script fails
    --arg ARG       Pass ARG to each script
    -h, --help      Show this help message

Examples:
    run-parts /etc/cron.daily       # Run all daily cron jobs
    run-parts --test /etc/cron.d    # Show what would be run
    run-parts -v /usr/local/bin     # Run with verbose output
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the run-parts command."""
        if handle_help_flag(self, args):
            return 0

        # Parse flags
        parse_result = parse_flags(
            args,
            {
                "test": bool,
                "list": bool,
                "v": bool,
                "verbose": bool,
                "report": bool,
                "exit-on-error": bool,
                "arg": str,
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        if not positional_args:
            self.console.print("[red]run-parts: missing directory[/red]")
            return 1

        directory = positional_args[0]
        test_mode = flags.get("test", False) or flags.get("list", False)
        verbose = flags.get("v", False) or flags.get("verbose", False)
        report = flags.get("report", False)
        exit_on_error = flags.get("exit-on-error", False)
        extra_arg = flags.get("arg")

        try:
            # List files in the directory
            try:
                files = client.list_files(directory)
            except Exception as e:
                self.console.print(
                    f"[red]run-parts: cannot read directory '{directory}': {e}[/red]"
                )
                return 1

            # Filter and sort executable files (or treat all as potentially executable)
            executable_files = [
                file_info.name
                for file_info in files
                if file_info.type == ops.pebble.FileType.FILE
                and self._is_valid_run_parts_file(file_info.name)
            ]

            # Sort lexicographically
            executable_files.sort()

            if not executable_files:
                if verbose:
                    self.console.print(
                        f"run-parts: no executable files found in {directory}"
                    )
                return 0

            # Test mode - just list files
            if test_mode:
                for filename in executable_files:
                    file_path = f"{directory.rstrip('/')}/{filename}"
                    self.console.print(file_path)
                return 0

            # Execute each file
            overall_exit_code = 0
            for filename in executable_files:
                file_path = f"{directory.rstrip('/')}/{filename}"

                if verbose or report:
                    self.console.print(f"run-parts: executing {file_path}")

                try:
                    # Build command arguments
                    command = [file_path]
                    if extra_arg:
                        command.append(extra_arg)

                    # Execute the script
                    process = client.exec(
                        command,
                        working_dir="/",
                        user_id=0,
                        user="root",
                        group_id=0,
                        group="root",
                    )

                    # Stream output
                    for line in process.stdout:
                        self.console.print(line, end="")

                    # Wait for completion
                    process.wait()
                    exit_code = process.exit_code or 0

                    if exit_code != 0:
                        overall_exit_code = exit_code
                        if verbose:
                            self.console.print(
                                f"[yellow]run-parts: {file_path} exited with code {exit_code}[/yellow]"
                            )

                        if exit_on_error:
                            self.console.print(
                                f"[red]run-parts: stopping due to error in {file_path}[/red]"
                            )
                            return exit_code

                except Exception as e:
                    self.console.print(
                        f"[red]run-parts: failed to execute {file_path}: {e}[/red]"
                    )
                    overall_exit_code = 1
                    if exit_on_error:
                        return 1

            return overall_exit_code

        except Exception as e:
            self.console.print(f"[red]run-parts: {e}[/red]")
            return 1

    def _is_valid_run_parts_file(self, filename: str) -> bool:
        """Check if filename is valid for run-parts execution."""
        # run-parts typically ignores files with certain patterns
        # Skip files that start with . or end with ~ or contain ,
        if filename.startswith("."):
            return False
        if filename.endswith("~"):
            return False
        if "," in filename:
            return False
        if filename.endswith(".disabled"):
            return False
        if filename.endswith(".dpkg-old"):
            return False
        if filename.endswith(".dpkg-new"):
            return False
        if filename.endswith(".rpmsave"):
            return False
        if filename.endswith(".rpmnew"):
            return False

        # Only allow alphanumeric, underscore, hyphen, and dot
        allowed_chars = string.ascii_letters + string.digits + "_-."
        return all(c in allowed_chars for c in filename)



"""Integration tests for the XargsCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create an XargsCommand instance."""
    yield pebble_shell.commands.XargsCommand(shell=shell)


def test_name(command: pebble_shell.commands.XargsCommand):
    assert command.name == "xargs"


def test_category(command: pebble_shell.commands.XargsCommand):
    assert command.category == "Advanced Utilities"


def test_help(command: pebble_shell.commands.XargsCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "execute" in capture.get() or "arguments" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "xargs" in capture.get() or "execute" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=[])
    # xargs without arguments should succeed using default command
    assert result == 0


def test_execute_with_command(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["echo"])
    # Should succeed executing echo command
    assert result == 0


def test_execute_with_max_args_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-n", "2", "echo"])
    # Should succeed executing echo command
    assert result == 0


def test_execute_with_delimiter_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-d", "\n", "echo"])
    # Should succeed executing echo command
    assert result == 0


def test_execute_with_null_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-0", "echo"])
    # Should succeed executing echo command
    assert result == 0


def test_execute_with_interactive_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-p", "echo"])
    # Should succeed executing echo command
    assert result == 0


def test_execute_with_complex_command(
    client: ops.pebble.Client,
    command: pebble_shell.commands.XargsCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-n", "1", "ls", "-la"])
    # Should succeed executing echo command
    assert result == 0



"""Remove files and directories command."""

from __future__ import annotations

from typing import TYPE_CHECKING

from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn

if TYPE_CHECKING:
    import ops
    import shimmer

from ...utils import expand_globs_in_tokens, resolve_path
from ...utils.command_helpers import (
    handle_help_flag,
    parse_flags,
    validate_min_args,
)
from ...utils.file_ops import (
    remove_file_recursive,
)
from .._base import Command


class RemoveCommand(Command):
    """Remove files and directories."""

    name = "rm"
    help = "Remove files and directories"
    category = "Filesystem Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute rm command."""
        if handle_help_flag(self, args):
            return 0

        if not validate_min_args(
            self.shell, args, 1, "rm [-r] [-f] <file1> [file2...]"
        ):
            return 1

        flags_result = parse_flags(args, {"r": bool, "f": bool}, self.shell)
        if flags_result is None:
            return 1

        flags, remaining_args = flags_result
        force = flags.get("f", False)

        if not remaining_args:
            self.console.print("rm: missing operand")
            return 1

        # Expand globs in file arguments
        expanded_files: list[str] = []
        for file_pattern in remaining_args:
            expanded = expand_globs_in_tokens(
                client, [file_pattern], self.shell.current_directory
            )
            if expanded:
                expanded_files.extend(expanded)
            elif not force:
                self.console.print(f"rm: No files match pattern '{file_pattern}'")
                return 1

        if not expanded_files:
            if not force:
                self.console.print("rm: No files to remove")
                return 1
            return 0

        files = [
            resolve_path(self.shell.current_directory, f, self.shell.home_dir)
            for f in expanded_files
        ]

        exit_code = 0
        with Progress(
            SpinnerColumn(),
            BarColumn(),
            TextColumn("{task.description}"),
            transient=True,
        ) as progress:
            task = progress.add_task("Removing files...", total=len(files))
            for file_path in files:
                # Use file_ops utility for remove operation
                success = remove_file_recursive(
                    client, self.console, file_path, force, progress, task
                )
                if not success and not force:
                    exit_code = 1

        return exit_code



"""Implementation of UncompressCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from .._base import Command
from .compress import CompressCommand

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class UncompressCommand(Command):
    """Implementation of uncompress command."""

    name = "uncompress"
    help = "Decompress files compressed with compress"
    category = "Compression"

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute uncompress command."""
        compress_cmd = CompressCommand(self.shell)
        return compress_cmd.execute(client, ["-d", *args])



"""Head command for Cascade."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence


from ._base import _LinesCommand


class HeadCommand(_LinesCommand):
    """Display first lines of a file."""

    name = "head"
    help = "Display first lines of file"

    def process_lines(self, file_lines: Sequence[str], lines: int):
        """Display first lines of a file."""
        for line in file_lines[:lines]:
            self.shell.console.print(line)
